from starlette.applications import Starlette
from starlette.testclient import TestClient

from starlette_jsonrpc import dispatcher
from starlette_jsonrpc.endpoint import JSONRPCEndpoint


app = Starlette()


@dispatcher.add_method
async def subtract(params):
    return {"result": params["x"] - params["y"]}


@dispatcher.add_method(name="SubtractMethod")
async def second_method(params):
    return {"result": params["x"] - params["y"]}


@dispatcher.add_method
async def subtract_positional(x, y):
    return {"result": x - y}


app.mount("/api", JSONRPCEndpoint)

client = TestClient(app)



import typesystem


class ErrorSchema(typesystem.Schema):
    code = typesystem.Integer()
    message = typesystem.String()
    data = typesystem.Object(additional_properties=True)


class JSONRPCRequest(typesystem.Schema):
    jsonrpc = typesystem.String(pattern="2.0", trim_whitespace=False)
    id = typesystem.Union(
        any_of=[
            typesystem.String(allow_null=True, min_length=1, trim_whitespace=False),
            typesystem.Integer(allow_null=True),
        ]
    )
    params = typesystem.Union(
        any_of=[typesystem.Object(additional_properties=True), typesystem.Array()]
    )
    method = typesystem.String()


class JSONRPCResponse(typesystem.Schema):
    jsonrpc = typesystem.String(pattern="2.0", trim_whitespace=False)
    id = typesystem.Union(
        any_of=[
            typesystem.String(allow_null=True, min_length=1, trim_whitespace=False),
            typesystem.Integer(allow_null=True),
        ]
    )
    result = typesystem.Object()


class JSONRPCErrorResponse(typesystem.Schema):
    jsonrpc = typesystem.String(pattern="2.0", trim_whitespace=False)
    id = typesystem.String(trim_whitespace=False)
    error = typesystem.Reference(to=ErrorSchema)



from . import client


# JSON


def test_payload_as_empty_dict():
    payload = {}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": 'None',
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {},
        },
    }


def test_payload_as_empty_list():
    payload = []
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": 'None',
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {},
        },
    }


def test_incorrect_payload():
    payload = [1]
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": 'None',
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {},
        },
    }


# PARAMS


def test_positional_parameters():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract_positional",
        "params": [42, 23],
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": {"result": 19}}


def test_named_parameters():
    payload = {
        "jsonrpc": "2.0",
        "method": "SubtractMethod",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": {"result": 19}}


def test_params_not_object():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": "", "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"params": "Did not match any valid type."},
        },
    }


def test_params_as_invalid_object():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"params": "Required param: 'x'"},
        },
    }


def test_params_as_invalid_list():
    payload = {"jsonrpc": "2.0", "method": "subtract_positional", "params": [1, ], "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"params": "subtract_positional() missing 1 required positional argument: 'y'"},
        },
    }


def test_without_params():
    payload = {"jsonrpc": "2.0", "method": "my_method", "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.status_code == 200


# ID

def test_id_as_integer():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {"x": 42, "y": 23}, "id": 1}
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": 1, "result": {"result": 19}}


def test_id_as_string():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {"x": 42, "y": 23}, "id": "abc"}
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "abc", "result": {"result": 19}}


def test_id_as_null():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {"x": 42, "y": 23}, "id": None}
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": None, "result": {"result": 19}}


def test_empty_id():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {"x": 42, "y": 23}, "id": ""}
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": None, "result": {"result": 19}}


def test_without_id():
    """
    Notification
    """
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {"x": 42, "y": 23}}
    response = client.post("/api/", json=payload)
    assert response.json() == {}


# JSONRPC


def test_jsonrpc_as_integer():
    payload = {"jsonrpc": 2, "method": "subtract", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "Must be a string."},
        },
    }


def test_empty_jsonrpc():
    payload = {"jsonrpc": "", "method": "subtract", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "Must not be blank."},
        },
    }


def test_jsonrpc_wrong_value():
    payload = {"jsonrpc": "3.0", "method": "subtract", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "Must match the pattern /2.0/."},
        },
    }


def test_without_jsonrpc():
    payload = {"method": "subtract", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "This field is required."},
        },
    }


# METHOD


def test_not_registered_method():
    payload = {"jsonrpc": "2.0", "method": "non_existing_method", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {"code": -32601, "message": "Method not found.", "data": {}},
    }


def test_without_method():
    payload = {"jsonrpc": "2.0", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"method": "This field is required."},
        },
    }


def test_with_empty_method():
    payload = {"jsonrpc": "2.0", "method": "", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"method": "Must not be blank."},
        },
    }


def test_method_as_integer():
    payload = {"jsonrpc": "2.0", "method": 1, "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"method": "Must be a string."},
        },
    }


# def test_with_method_name_starting_with_rpc_period():
#     pass



from . import client


def test_post_call_should_return_status_code_200():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.status_code == 200


def test_post_for_named_function_should_return_status_code_200():
    payload = {
        "jsonrpc": "2.0",
        "method": "SubtractMethod",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.status_code == 200


def test_get_call_should_return_method_not_allowed():
    response = client.get("/api/")
    assert response.status_code == 405


def test_put_call_should_return_method_not_allowed():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.put("/api/", data=payload)
    assert response.status_code == 405


def test_delete_call_should_return_method_not_allowed():
    response = client.delete("/api/")
    assert response.status_code == 405


def test_patch_call_should_return_method_not_allowed():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.patch("/api/", payload)
    assert response.status_code == 405



import functools


class Dispatcher:
    def __init__(self):
        self.routes_map = {}

    def add_method(self, method=None, name=None):

        if name and not method:
            return functools.partial(self.add_method, name=name)

        self.routes_map[name or method.__name__] = method
        return method



import os
import setuptools


dir_path = os.path.dirname(os.path.realpath(__file__))


def read(fname):
    with open(fname, "r") as f:
        return f.read()


install_requires = [
    l
    for l in read(os.path.join(dir_path, "requirements.txt")).splitlines()
    if l and not l.startswith("#")
]


setuptools.setup(
    name="starlette-jsonrpc",
    version="0.0.0",
    author="Kamil Dębowski",
    author_email="poczta@kdebowski.pl",
    description="JSON-RPC implementation for Starlette framework",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/kdebowski/starlette-jsonrpc",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    install_requires=install_requires,
    include_package_data=True,
)
